/* 
 * File:   main.c
 * Author: danjulio
 *
 * DVM for 12V system and LCD display.  Displays XX.X on a
 * PD-422 display.  Uses a regulated 5V input as reference.
 * 
 *   AN2 - DMM Input scaled by 1/3
 * 
 *   LCD Segment assignment
 *     SEG6:0  - Fractional LCD Segments G:A
 *     SEG7    - Decimal Point (always enabled)
 *     SEG14:8 - Units LCD Segments G:A
 *     SEG15   - Tens LCD Segments B & C (wired together to make "1")
 *     RC2     - VLCD3 - 5V Bias
 * 
 *   RB5 - Toggling heartbeat output
 * 
 * Takes a reading about every 125 mSec and averages in a 16-entry buffer.
 * 
 */
#include <xc.h>
#include <stdio.h>
#include <stdint.h>

// Config fuse bits
#pragma config DEBUG = OFF
#pragma config FCMEN = OFF
#pragma config IESO = 0
#pragma config BOREN = OFF
#pragma config CPD = OFF
#pragma config CP = OFF
#pragma config MCLRE = 0
#pragma config PWRTE = ON
#pragma config WDTE = OFF
#pragma config FOSC = INTOSCIO

// Calibration value - per device
//   Per Device Error (PDE) = Percentage error in reading (accounting for resistor divider & 5V PSU ref)
//      PDE = Error from resistors + Error from voltage regulator
//          Error from resistors is measured by applying a known voltage and measuring the output
//             100 * ((Vmeas - (Vin/3)) / (Vin/3))
//          Error from 5V regulator is measured
//             100 * (5 - Vmeas) / 5)
//   Scale Factor (SF) = Percent amount to adjust reading by to calibrate = 100 - PDE
//   Cal Value (CV) = SF / 100
//
//   We apply the SF by adjusting the divisor with the CV
//      V = (ADCcount * 15 * CV) / 1023 = (ADCcount * 15) / (1023 * 1/CV)
//      V = (ADCcount * 15) / CAL_DIV
//    
//   CAL_DIV = 1023 * 1/CV
// 
//   Default CAL_DIV = 1023
#define CAL_DIV 1023

// LCD lookup tables
const uint8_t fracTable[] ={0x3F, 0x06, 0x5B, 0x4F, 0x66, 0x6D, 0x7D, 0x07, 0x7F, 0x6F};
const uint8_t intTable[] = {0x3F, 0x06, 0x5B, 0x4F, 0x66, 0x6D, 0x7D, 0x07, 0x7F, 0x6F,
                            0xBF, 0x86, 0xDB, 0xCF, 0xE6, 0xED};


// Variables
uint16_t avgBuff[16];
uint8_t avgIndex = 0;
uint16_t avgVal;


void InitPic() {
    // Configure watchdog timeout for 31 kHz oscillator
    OPTION_REGbits.PSA = 0;
    WDTCON = 0x13;      // 1:16384 (~500 mSec)
    CLRWDT();
    
    // Configure unused digital IO Pins to output 0
    TRISA = 0x3F;  // RA[7:6] outputs
    TRISB = 0xDF;  // RB[5] output
    TRISC = 0xFC;  // RC[1:0] outputs
    
    // Configure clock for 1 MHz
    OSCCON = 0x41;
    
    // Configure ADC for AN2 input, right-justified output
    ADCON0 = 0x89;  // internal Vref
    ADCON1 = 0x40;  // FOSC/4 => 4 uSec conversion time
    
    // Configure LCD for single-phase, internal 31 kHz oscillator
    LCDPS = 0x03;          // ~ 60 Hz refresh
    LCDSE0 = 0xFF;
    LCDSE1 = 0xFF;
    LCDCON = 0x08;         // Set LMUX = 00, CS = 10, SLPENn = 0
    LCDDATA0 = 0xBF;       // "0.0"
    LCDDATA1 = 0x3F;
    PIR2bits.LCDIF = 0;
    LCDCONbits.VLCDEN = 1; // Enable bias voltage pins
    LCDCONbits.LCDEN = 1;  // Enable LCD Module
    
    // Configure Timer2 to interrupt about every 125 mSec
    //   Prescale 1:16  => 64 uSec count
    //   PR2 = 122      => 7.808 mSec counter roll-over
    //   Postscale 1:16 => 124.928 mSec timeout
    T2CON = 0x7E;
    PR2 = 122;
    PIR1bits.TMR2IF = 0;
}


void GetAdc() {
    uint8_t i;
    
    // Trigger the ADC
    ADCON0bits.GO = 1;
    while (ADCON0bits.GO == 1) {};
    
    // Push this reading
    avgBuff[avgIndex++] = ((ADRESH & 0x03) << 8) | ADRESL;
    if (avgIndex == 16) avgIndex = 0;
    
    // Compute the current average of readings
    avgVal = 0;
    for (i=0; i<16; i++) {
        avgVal += avgBuff[i];
    }
    // We can round on avgVal[3]
    if ((avgVal & 0x0008) != 0) {
        avgVal += 0x0010;
    }
    // After rounding, we can discard the fractional part
    avgVal = avgVal >> 4;
}


void UpdateLcd() {
    uint16_t sV;
    uint8_t intVal;
    uint8_t fracVal;
    
    // Compute the scaled value that represents our fractional voltage
    //   1. Multiply ADC count by 15 (3 voltage scale factor * 5 volts vref)
    //   2. Divide by CAL_DIV to get the integer portion of volts
    //   3. Take the modulo with CAL_DIV to get the fractional portion of volts
    //   4. Convert the fractional volts to a tenths value
    //      a. Multiply fractional volts by 10
    //         Result has the form:
    //            Bits 15:10 - Tenths value (0 - 9 in bits 13:10)
    //            Bits  9: 0 - Fraction
    //      b. Round Tenths value using bit 9.  Possibly round up the integer volts.
    //
    sV = avgVal * 15;
    intVal = sV / CAL_DIV;
    sV = sV % CAL_DIV;
    sV = sV * 10;
    fracVal = (sV >> 10) & 0x0F;
    if (sV & 0x0200) fracVal += 1;
    if (fracVal > 9) {
        fracVal = 0;
        intVal += 1;
    }
    
    // Compute the LCD segment data
    //
    // LCDDATA0 contains the tenths value and the decimal point
    // LCDDATA1 contains the units value (bit 7 is the tens)
    fracVal = fracTable[fracVal] | 0x80;
    intVal = intTable[intVal];
    
    // Wait until its ok to update the LCD Module
    while (LCDPSbits.WA == 0) {};
    LCDDATA0 = fracVal;
    LCDDATA1 = intVal;
}



/*
 * MAIN
 */
int main() {
    uint8_t i;
    
    InitPic();
    
    // Get an initial reading
    for (i=0; i<16; i++) {
        GetAdc();
    }
    UpdateLcd();
    
    // Loop forever, updating the voltage display
    while (1) {
        CLRWDT();
        
        // Wait for TIMER2 to roll-over
        while (PIR1bits.TMR2IF == 0) {};
        PIR1bits.TMR2IF = 0;
        
        GetAdc();
        UpdateLcd();
        
        // Let them know we're alive
        PORTBbits.RB5 = ~PORTBbits.RB5;
    }

    return (0);
}

